@@ -7,13 +7,14 @@ import gc
 import logging
 import os
 from pathlib import Path
+import re
 from tempfile import TemporaryDirectory
-from typing import Dict, List, Optional, Union
-from frads.utils import parse_polygon, parse_rad_header, random_string
-from frads.geom import Polygon
+
+from frads.geom import parse_polygon, polygon_primitive, Polygon
+from frads.utils import random_string
 import numpy as np
-from scipy.sparse import csr_matrix, lil_matrix
 import pyradiance as pr
+from scipy.sparse import csr_matrix, lil_matrix
 
 
 logger: logging.Logger = logging.getLogger("frads.matrix")
@@ -45,7 +46,7 @@ class SensorSender:
         yres: Number of sensors
     """
 
-    def __init__(self, sensors: List[List[float]], ray_count: int = 1):
+    def __init__(self, sensors: list[list[float]], ray_count: int = 1):
         """Initialize a sender object with a list of sensors:
 
         Args:
@@ -77,19 +78,20 @@ class ViewSender:
         self.view = view
         self.xres = xres
         self.yres = yres
+        view_args = pr.get_view_args(view)
         res_eval = pr.vwrays(
-            view=view.args(), xres=xres, yres=yres, dimensions=True
+            view=view_args, xres=xres, yres=yres, dimensions=True
         ).split()
         new_xres, new_yres = int(res_eval[1]), int(res_eval[3])
         vwrays_proc = pr.vwrays(
-            view=view.args(),
+            view=view_args,
             outform="f",
             xres=new_xres,
             yres=new_yres,
             ray_count=ray_count,
             pixel_jitter=0.7,
         )
-        if view.vtype == "a":
+        if view.type == "a":
             ray_flush_exp = (
                 f"DIM:{xres};CNT:{ray_count};"
                 "pn=(recno-1)/CNT+.5;"
@@ -125,7 +127,7 @@ class SurfaceSender:
 
     def __init__(
         self,
-        surfaces: List[pr.Primitive],
+        surfaces: list[pr.Primitive],
         basis: str,
         left_hand: bool = False,
         offset: bool = None,
@@ -138,10 +140,10 @@ class SurfaceSender:
             left_hand: Whether to use left-hand coordinate system
             offset: Offset of the sender surface
         """
-        self.surfaces = surfaces
+        self.surfaces = [s for s in surfaces if s.ptype in ("polygon", "ring")]
         self.basis = basis
         self.content = rfluxmtx_markup(
-            surfaces, basis, left_hand=left_hand, offset=offset
+            self.surfaces, basis, left_hand=left_hand, offset=offset
         )
 
     def __eq__(self, other):
@@ -183,7 +185,7 @@ class SurfaceReceiver(Receiver):
 
     def __init__(
         self,
-        surfaces: List[pr.Primitive],
+        surfaces: list[pr.Primitive],
         basis: str,
         left_hand=False,
         offset=None,
@@ -193,9 +195,9 @@ class SurfaceReceiver(Receiver):
         super().__init__(basis)
         if not isinstance(surfaces[0], pr.Primitive):
             raise ValueError("Surface must be a primitive", surfaces)
-        self.surfaces = surfaces
+        self.surfaces = [s for s in surfaces if s.ptype in ("polygon", "ring")]
         self.content = rfluxmtx_markup(
-            surfaces,
+            self.surfaces,
             basis=basis,
             left_hand=left_hand,
             offset=offset,
@@ -218,12 +220,12 @@ class SkyReceiver(Receiver):
         if out is not None:
             self.content += f"#@rfluxmtx o={out}\n"
         self.content += (
-            f"#@rfluxmtx h={basis} u=+Y\n"
-            "void glow skyglow 0 0 4 1 1 1 0\n"
-            "skyglow source skydome 0 0 4 0 0 1 180\n"
             f"#@rfluxmtx h=u\n"
             "void glow groundglow 0 0 4 1 1 1 0\n"
             "groundglow source groundplane 0 0 4 0 0 -1 180\n"
+            f"#@rfluxmtx h={basis} u=+Y\n"
+            "void glow skyglow 0 0 4 1 1 1 0\n"
+            "skyglow source skydome 0 0 4 0 0 1 180\n"
         )
 
 
@@ -243,8 +245,8 @@ class SunReceiver(Receiver):
     def __init__(
         self,
         basis: str,
-        sun_matrix: Optional[np.ndarray] = None,
-        window_normals: Optional[List[np.ndarray]] = None,
+        sun_matrix: None | np.ndarray = None,
+        window_normals: None | list[np.ndarray] = None,
         full_mod: bool = False,
     ):
         """Initialize a sun receiver.
@@ -299,10 +301,10 @@ class Matrix:
 
     def __init__(
         self,
-        sender: Union[SensorSender, ViewSender, SurfaceSender],
-        receivers: Union[List[SkyReceiver], List[SurfaceReceiver], List[SunReceiver]],
-        octree: Optional[str] = None,
-        surfaces: Optional[List[str]] = None,
+        sender: SensorSender | ViewSender | SurfaceSender,
+        receivers: list[SkyReceiver] | list[SurfaceReceiver] | list[SunReceiver],
+        octree: None | str = None,
+        surfaces: None | list[pr.Primitive] = None,
     ):
         """Initialize a matrix.
 
@@ -345,7 +347,7 @@ class Matrix:
 
     def generate(
         self,
-        params: List[str],
+        params: list[str],
         nproc: int = 1,
         sparse: bool = False,
         to_file: bool = False,
@@ -386,6 +388,12 @@ class Matrix:
             receiver_file = os.path.join(tmpdir, "receiver")
             with open(receiver_file, "w") as f:
                 [f.write(r.content) for r in self.receivers]
+            env_file = None
+            if self.surfaces is not None:
+                env_file = os.path.join(tmpdir, "scene")
+                with open(env_file, "wb") as f:
+                    f.write(b" ".join(s.bytes for s in self.surfaces))
+                env_file = [env_file]
             if isinstance(self.sender, SurfaceSender):
                 params.append("-ffd")
                 surface_file = os.path.join(tmpdir, "surface")
@@ -397,7 +405,7 @@ class Matrix:
                 rays=rays,
                 params=params,
                 octree=self.octree,
-                scene=self.surfaces,
+                scene=env_file,
             )
         if not to_file:
             _ncols = sum(self.ncols) if isinstance(self.ncols, list) else self.ncols
@@ -469,10 +477,10 @@ class SunMatrix(Matrix):
 
     def __init__(
         self,
-        sender: Union[SensorSender, ViewSender],
+        sender: SensorSender | ViewSender,
         receiver: SunReceiver,
-        octree: Optional[str],
-        surfaces: Optional[List[str]] = None,
+        octree: None | str,
+        surfaces: None | list[str] = None,
     ):
         if isinstance(sender, SurfaceSender):
             raise TypeError("SurfaceSender cannot be used with SunMatrix")
@@ -483,11 +491,11 @@ class SunMatrix(Matrix):
 
     def generate(
         self,
-        parameters: List[str],
+        parameters: list[str],
         nproc: int = 1,
         radmtx: bool = False,
         sparse: bool = False,
-    ) -> Optional[bytes]:
+    ) -> None | bytes:
         """Call rcontrib to generate a matrix and store it in memory
         as a numpy array.
 
@@ -575,7 +583,30 @@ class SunMatrix(Matrix):
             )
 
 
-def load_matrix(file: Union[bytes, str, Path], dtype: str = "float") -> np.ndarray:
+def parse_rad_header(header_str: str) -> tuple:
+    """Parse a Radiance matrix file header.
+
+    Args:
+        header_str: header as string
+    Returns:
+        A tuple contain nrow, ncol, ncomp, datatype
+    Raises:
+        ValueError: if any of NROWS NCOLS NCOMP FORMAT is not found.
+    """
+    compiled = re.compile(
+        r" NROWS=(.*) | NCOLS=(.*) | NCOMP=(.*) | FORMAT=(.*) ", flags=re.X
+    )
+    matches = compiled.findall(header_str)
+    if len(matches) != 4:
+        raise ValueError("Can't find one of the header entries.")
+    nrow = int([mat[0] for mat in matches if mat[0] != ""][0])
+    ncol = int([mat[1] for mat in matches if mat[1] != ""][0])
+    ncomp = int([mat[2] for mat in matches if mat[2] != ""][0])
+    dtype = [mat[3] for mat in matches if mat[3] != ""][0].strip()
+    return nrow, ncol, ncomp, dtype
+
+
+def load_matrix(file: bytes | str | Path, dtype: str = "float") -> np.ndarray:
     """Load a Radiance matrix file into numpy array.
 
     Args:
@@ -616,7 +647,7 @@ def load_binary_matrix(
 
 
 def matrix_multiply_rgb(
-    *mtx: np.ndarray, weights: Optional[List[float]] = None
+    *mtx: np.ndarray, weights: None | list[float] = None
 ) -> np.ndarray:
     """Multiply matrices as numpy ndarray.
     linalg.multi_dot figures out the multiplication order
@@ -644,7 +675,7 @@ def sparse_matrix_multiply_rgb_vtds(
     tmx: np.ndarray,
     dmx: np.ndarray,
     smx: np.ndarray,
-    weights: Optional[List[float]] = None,
+    weights: None | list[float] = None,
 ) -> np.ndarray:
     """Multiply sparse view, transmission, daylight,
     and sky matrices. (ThreePhaseMethod)
@@ -703,12 +734,12 @@ def to_sparse_matrix3(mtx: np.ndarray, mtype: str = "csr") -> np.ndarray:
 
 
 def rfluxmtx_markup(
-    surfaces: List[pr.Primitive],
+    surfaces: list[pr.Primitive],
     basis: str,
     left_hand: bool = False,
-    offset: Optional[float] = None,
+    offset: None | float = None,
     source: str = "glow",
-    out: Optional[str] = None,
+    out: None | str = None,
 ) -> str:
     """Mark up surfaces for rfluxmtx.
 
@@ -724,7 +755,7 @@ def rfluxmtx_markup(
         Marked up primitives as strings (to be written to a file for rfluxmtx)
     """
     modifier_set = {p.modifier for p in surfaces}
-    source_modifier = f"rflx{surfaces[0].modifier}{random_string(5)}"
+    source_modifier = f"rflx{surfaces[0].modifier}{random_string(10)}"
     if left_hand:
         basis = "-" + basis
     if source not in ("glow", "light"):
@@ -782,15 +813,15 @@ def rfluxmtx_markup(
 
 
 def surfaces_view_factor(
-    surfaces: List[pr.Primitive],
-    env: List[pr.Primitive],
+    surfaces: list[pr.Primitive],
+    env: list[pr.Primitive],
     ray_count: int = 10000,
-) -> Dict[str, List[float]]:
+) -> dict[str, dict[str, list[float]]]:
     """Calculate surface to surface view factor using rfluxmtx.
 
     Args:
         surfaces: list of surface primitives that we want to calculate view factor for.
-            Surface normal needs to be facing inward.
+            Surface normal needs to be facing outward.
         env: list of environment primitives that our surfaces will be exposed to.
             Surface normal needs to be facing inward.
         ray_count: number of rays spawned for each surface.
@@ -799,11 +830,23 @@ def surfaces_view_factor(
         A dictionary of view factors, where the key is the surface identifier.
     """
     view_factors = {}
-    surfaces_env = surfaces + env
-    receivers = [SurfaceReceiver([s], basis="u") for s in surfaces_env]
-    for surface in surfaces:
+    surfaces_env = env + surfaces
+    for idx, surface in enumerate(surfaces):
         sender = SurfaceSender([surface], basis="u")
+        rest_of_surfaces = surfaces[:idx] + surfaces[idx + 1 :]
+        rest_of_surface_polygons = [parse_polygon(s) for s in rest_of_surfaces]
+        rest_of_surfaces_flipped = [
+            polygon_primitive(p.flip(), s.modifier, s.identifier)
+            for s, p in zip(rest_of_surfaces, rest_of_surface_polygons)
+        ]
+        receivers = [
+            SurfaceReceiver([s], basis="u")
+            for s in env + rest_of_surfaces_flipped
+            if s.ptype in ("polygon", "ring")
+        ]
         mat = Matrix(sender, receivers, octree=None, surfaces=surfaces_env)
         mat.generate(params=["-c", f"{ray_count}"])
-        view_factors[surface.identifier] = [[i][0] for i in mat.array]
+        view_factors[surface.identifier] = {
+            r.surfaces[0].identifier: [i][0] for r, i in zip(mat.receivers, mat.array)
+        }
     return view_factors
